/* File:		draw_terrain.c */
/* Description:	This file module contains the terrain drawing functions */

#include <stdio.h>
#include <math.h>
#include <GL/glut.h>
#include "types.h"
#include "maths.h"
#include "terrain.h"

/* variables */

extern world_vector position;
extern orientation angle;

/* Function to generate grid distances for current viewer position */

void calculate_distance_map()
{
	int viewer_s, viewer_t;
	world_vector vector;
	dir_cosine dc;
	orientation a;
	int s, t;

	/* find grid position for viewer */

	position_to_grid_coord(position, &viewer_s, &viewer_t);

	/* calculate current view matrix */

	a.yaw = angle.yaw;
	a.pitch = 0.0;
	a.roll = 0.0;
	dc = calculate_dir_cosines(a);

	/* find distances to viewer for each square in the grid */

	for (s=0; s<257; s++)
	{
		for (t=0; t<257; t++)
		{
			vector.x = s - viewer_s;
			vector.y = 0.0;
			vector.z = t - viewer_t;
			vector = transform_world_to_view(vector, dc);
			distancemap[s][t].s = vector.x;
			distancemap[s][t].t = vector.z;
		}
	}
}

/* Function to determine if a position in the terrain is visible */

int terrain_visible(int s, int t)
{
	float yaw_diff;

	/* check if square is behind viewer */
	
	if (distancemap[s][t].t < -20.0)
		return 0;
	if (distancemap[s][t].t < 20.0 && (distancemap[s][t].s > -20.0 && distancemap[s][t].s < 20.0))
		return 1;
	yaw_diff = atan2d(distancemap[s][t].s,distancemap[s][t].t);

	/* determine if visible */

	if (yaw_diff >= -38.0 && yaw_diff <= 38.0)
		return 1;

	return 0;
}

/* Function to cull parent cells from the scene */

void cull_cells()
{
	int n, m;
	int sl, tb;					/* grid coords for bottom left corner of cell */
	int sc, tc;					/* grid coords for center of cell */
	int sr, tt;					/* grid coord for top right of cell */

	for (n=0; n<8; n++)
	{
		for (m=0; m<8; m++)
		{
			sl = lod5[n][m].s_datum;
			tb = lod5[n][m].t_datum;
			sr = lod5[n][m].s_datum + lod5[n][m].grid_step;
			tt = lod5[n][m].t_datum + lod5[n][m].grid_step;
			if (terrain_visible(sl, tb) || terrain_visible(sr, tb) ||
				terrain_visible(sr, tt) || terrain_visible(sl, tt))
			{
				lod5[n][m].flags = 0x8000;
			}
			else
			{
				lod5[n][m].flags = 0;
			}
		}
	}
	for (n=0; n<16; n++)
	{
		for (m=0; m<16; m++)
		{
			sl = lod4[n][m].s_datum;
			tb = lod4[n][m].t_datum;
			sr = lod4[n][m].s_datum + lod4[n][m].grid_step;
			tt = lod4[n][m].t_datum + lod4[n][m].grid_step;
			if (terrain_visible(sl, tb) || terrain_visible(sr, tb) ||
				terrain_visible(sr, tt) || terrain_visible(sl, tt))
			{
				lod4[n][m].flags = 0x8000;
			}
			else
			{
				lod4[n][m].flags = 0;
			}
		}
	}
	for (n=0; n<32; n++)
	{
		for (m=0; m<32; m++)
		{
			sl = lod3[n][m].s_datum;
			tb = lod3[n][m].t_datum;
			sr = lod3[n][m].s_datum + lod3[n][m].grid_step;
			tt = lod3[n][m].t_datum + lod3[n][m].grid_step;
			if (terrain_visible(sl, tb) || terrain_visible(sr, tb) ||
				terrain_visible(sr, tt) || terrain_visible(sl, tt))
			{
				lod3[n][m].flags = 0x8000;
			}
			else
			{
				lod3[n][m].flags = 0;
			}
		}
	}
	for (n=0; n<64; n++)
	{
		for (m=0; m<64; m++)
		{
			sl = lod2[n][m].s_datum;
			tb = lod2[n][m].t_datum;
			sr = lod2[n][m].s_datum + lod2[n][m].grid_step;
			tt = lod2[n][m].t_datum + lod2[n][m].grid_step;
			if (terrain_visible(sl, tb) || terrain_visible(sr, tb) ||
				terrain_visible(sr, tt) || terrain_visible(sl, tt))
			{
				lod2[n][m].flags = 0x8000;
			}
			else
			{
				lod2[n][m].flags = 0;
			}
		}
	}
	for (n=0; n<128; n++)
	{
		for (m=0; m<128; m++)
		{
			sl = lod1[n][m].s_datum;
			tb = lod1[n][m].t_datum;
			sr = lod1[n][m].s_datum + lod1[n][m].grid_step;
			tt = lod1[n][m].t_datum + lod1[n][m].grid_step;
			if (terrain_visible(sl, tb) || terrain_visible(sr, tb) ||
				terrain_visible(sr, tt) || terrain_visible(sl, tt))
			{
				lod1[n][m].flags = 0x8000;
			}
			else
			{
				lod1[n][m].flags = 0;
			}
		}
	}
}

/* Function to check the activation status of a cell's vertices */

void check_cell(terrain_cell *cell, terrain_cell *cellb, terrain_cell *cellr, 
				terrain_cell *cellt, terrain_cell *celll)
{
	int sl, tb;					/* grid coords for bottom left corner of cell */
	int sc, tc;					/* grid coords for center of cell */
	int sr, tt;					/* grid coord for top right of cell */
	float threshold;			/* maximum variance that will be tolerated */
	terrain_cell *ncell;		/* pointer to neighbour cell */

	/* set up grid coord data */

	sl = cell->s_datum;
	tb = cell->t_datum;
	sc = cell->s_datum + (cell->grid_step / 2);
	tc = cell->t_datum + (cell->grid_step / 2);
	sr = cell->s_datum + cell->grid_step;
	tt = cell->t_datum + cell->grid_step;

	/* check if bottom vertex should be enabled & check neighbours if so */

	threshold = ((float)distancemap[sc][tb].t / lod_factor);
	if ((cell->flags & 0x0800) || (cell->flags & 0x0400) ||
		(cell->flags & 0x0008) || ((float)cell->variance[0] > threshold))
	{
		cell->flags = cell->flags | 0x0008;
		if (cellb)
		{
			if (!(cellb->flags & 0x0002))
			{
				cellb->flags = cellb->flags | 0x0002;
			}
		}
	}

	/* check if right vertex should be enabled & check neighbours if so */

	threshold = ((float)distancemap[sc][tb].t / lod_factor);
	if ((cell->flags & 0x0400) || (cell->flags & 0x0200) ||
		(cell->flags & 0x0004) || ((float)cell->variance[0] > threshold))
	{
		cell->flags = cell->flags | 0x0004;
		if (cellr)
		{
			if (!(cellr->flags & 0x0001))
			{
				cellr->flags = cellr->flags | 0x0001;
			}
		}
	}

	/* check if top vertex should be enabled & check neighbours if so */

	threshold = ((float)distancemap[sc][tb].t / lod_factor);
	if ((cell->flags & 0x0200) || (cell->flags & 0x0100) ||
		(cell->flags & 0x0002) || ((float)cell->variance[0] > threshold))
	{
		cell->flags = cell->flags | 0x0002;
		if (cellt)
		{
			if (!(cellt->flags & 0x0008))
			{
				cellt->flags = cellt->flags | 0x0008;
			}
		}
	}

	/* check if left vertex should be enabled & check neighbours if so */

	threshold = ((float)distancemap[sc][tb].t / lod_factor);
	if ((cell->flags & 0x0100) || (cell->flags & 0x0800) ||
		(cell->flags & 0x0001) || ((float)cell->variance[0] > threshold))
	{
		cell->flags = cell->flags | 0x0001;
		if (celll)
		{
			if (!(celll->flags & 0x0004))
			{
				celll->flags = celll->flags | 0x0004;
			}
		}
	}
}

/* Function to calculate lod for current view */

void calculate_lod()
{
	int n, m;
	terrain_cell *cellb, *cellr, *cellt, *celll;
	int subarrays, subarrayt;	/* sub-cell array coords */

	for (n=0; n<128; n++)
	{
		for (m=0; m<128; m++)
		{
			if (lod1[n][m].flags & 0x8000)
			{
				cellb = cellr = cellt = celll = NULL;
				if (n > 0)
					celll = &lod1[n-1][m];
				if (n < 127)
					cellr = &lod1[n+1][m];
				if (m > 0)
					cellb = &lod1[n][m-1];
				if (m < 127)
					cellt = &lod1[n][m+1];
				check_cell(&lod1[n][m], cellb, cellr, cellt, celll);
			}
		}
	}
	for (n=0; n<64; n++)
	{
		for (m=0; m<64; m++)
		{
			if (lod2[n][m].flags & 0x8000)
			{
				subarrays = n * 2;
				subarrayt = m * 2;
				if (lod1[subarrays][subarrayt].flags & 0x000f)
					lod2[n][m].flags = lod2[n][m].flags | 0x0800;
				if (lod1[subarrays+1][subarrayt].flags & 0x000f)
					lod2[n][m].flags = lod2[n][m].flags | 0x0400;
				if (lod1[subarrays+1][subarrayt+1].flags & 0x000f)
					lod2[n][m].flags = lod2[n][m].flags | 0x0200;
				if (lod1[subarrays][subarrayt+1].flags & 0x000f)
					lod2[n][m].flags = lod2[n][m].flags | 0x0100;
				cellb = cellr = cellt = celll = NULL;
				if (n > 0)
					celll = &lod2[n-1][m];
				if (n < 63)
					cellr = &lod2[n+1][m];
				if (m > 0)
					cellb = &lod2[n][m-1];
				if (m < 63)
					cellt = &lod2[n][m+1];
				check_cell(&lod2[n][m], cellb, cellr, cellt, celll);
			}
		}
	}
	for (n=0; n<32; n++)
	{
		for (m=0; m<32; m++)
		{
			if (lod3[n][m].flags & 0x8000)
			{
				subarrays = n * 2;
				subarrayt = m * 2;
				if (lod2[subarrays][subarrayt].flags & 0x000f)
					lod3[n][m].flags = lod3[n][m].flags | 0x0800;
				if (lod2[subarrays+1][subarrayt].flags & 0x000f)
					lod3[n][m].flags = lod3[n][m].flags | 0x0400;
				if (lod2[subarrays+1][subarrayt+1].flags & 0x000f)
					lod3[n][m].flags = lod3[n][m].flags | 0x0200;
				if (lod2[subarrays][subarrayt+1].flags & 0x000f)
					lod3[n][m].flags = lod3[n][m].flags | 0x0100;
				cellb = cellr = cellt = celll = NULL;
				if (n > 0)
					celll = &lod3[n-1][m];
				if (n < 31)
					cellr = &lod3[n+1][m];
				if (m > 0)
					cellb = &lod3[n][m-1];
				if (m < 31)
					cellt = &lod3[n][m+1];
				check_cell(&lod3[n][m], cellb, cellr, cellt, celll);
			}
		}
	}
	for (n=0; n<16; n++)
	{
		for (m=0; m<16; m++)
		{
			if (lod4[n][m].flags & 0x8000)
			{
				subarrays = n * 2;
				subarrayt = m * 2;
				if (lod3[subarrays][subarrayt].flags & 0x000f)
					lod4[n][m].flags = lod4[n][m].flags | 0x0800;
				if (lod3[subarrays+1][subarrayt].flags & 0x000f)
					lod4[n][m].flags = lod4[n][m].flags | 0x0400;
				if (lod3[subarrays+1][subarrayt+1].flags & 0x000f)
					lod4[n][m].flags = lod4[n][m].flags | 0x0200;
				if (lod3[subarrays][subarrayt+1].flags & 0x000f)
					lod4[n][m].flags = lod4[n][m].flags | 0x0100;
				cellb = cellr = cellt = celll = NULL;
				if (n > 0)
					celll = &lod4[n-1][m];
				if (n < 15)
					cellr = &lod4[n+1][m];
				if (m > 0)
					cellb = &lod4[n][m-1];
				if (m < 15)
					cellt = &lod4[n][m+1];
				check_cell(&lod4[n][m], cellb, cellr, cellt, celll);
			}
		}
	}
	for (n=0; n<8; n++)
	{
		for (m=0; m<8; m++)
		{
			if (lod5[n][m].flags & 0x8000)
			{
				subarrays = n * 2;
				subarrayt = m * 2;
				if (lod4[subarrays][subarrayt].flags & 0x000f)
					lod5[n][m].flags = lod5[n][m].flags | 0x0800;
				if (lod4[subarrays+1][subarrayt].flags & 0x000f)
					lod5[n][m].flags = lod5[n][m].flags | 0x0400;
				if (lod4[subarrays+1][subarrayt+1].flags & 0x000f)
					lod5[n][m].flags = lod5[n][m].flags | 0x0200;
				if (lod4[subarrays][subarrayt+1].flags & 0x000f)
					lod5[n][m].flags = lod5[n][m].flags | 0x0100;
				cellb = cellr = cellt = celll = NULL;
				if (n > 0)
					celll = &lod5[n-1][m];
				if (n < 7)
					cellr = &lod5[n+1][m];
				if (m > 0)
					cellb = &lod5[n][m-1];
				if (m < 7)
					cellt = &lod5[n][m+1];
				check_cell(&lod5[n][m], cellb, cellr, cellt, celll);
			}
		}
	}
}

/* Function to draw sky dome */

void draw_sky()
{
	float r = 300.0;		/* radius of sky dome */
	int detail = 6;			/* detail level for sky dome */
 	float a, b;				/* angles */
	float step = 30.0;		/* angle step (180/detail) */
	float upper_band;		/* y co-ord of upper vertcies */
	float lower_band;		/* y co-ord of lower vertices */
	float ru, rl;			/* radii of upper and lower bands */
	float cu, cl;			/* colour factors for upper and lower bands */
	world_vector v[4];		/* panel vertices */
	world_vector n[4];		/* vertex normals */
	int p, m;				/* counters */

	glPushMatrix();
	glTranslatef((128*GRID_FACTOR), -100.0, -(128*GRID_FACTOR));
	glDisable(GL_TEXTURE_2D);
	glColor3f(0.5, 0.5, 1.0);
	for (p=0; p<(detail/2); p++)
	{
		a = p * step;
		upper_band = r * cosd(a);
		lower_band = r * cosd(a + step);
		ru = r * sind(a);
		rl = r * sind(a + step);
		cu = (1.0 - cosd(a)) * 0.3;
		cl = (1.0 - cosd(a + step)) * 0.3;
		for (m=0; m<(detail * 2); m++)
		{
			b = m * step;
			if (p == 0)		/* 1st band */
			{
				v[0].x = 0.0; v[0].y = r; v[0].z = 0.0;
				n[0].x = 0.0; n[0].y = 1.0; n[0].z = 0.0;
				v[1].x = rl * sind(b); v[1].y = lower_band; v[1].z = rl * cosd(b);
				n[1].x = v[1].x / r; n[1].y = v[1].y / r; n[1].z = v[1].z / r;
				v[2].x = rl * sind(b + step); v[2].y = lower_band; v[2].z = rl * cosd(b + step);
				n[2].x = v[2].x / r; n[2].y = v[2].y / r; n[2].z = v[2].z / r;
				glBegin(GL_TRIANGLES);
					glColor3f(0.5+cu, 0.5+cu, 1.0);
					glVertex3f(v[0].x, v[0].y, -v[0].z);
					glColor3f(0.5+cl, 0.5+cl, 1.0);
					glVertex3f(v[1].x, v[1].y, -v[1].z);
					glColor3f(0.5+cl, 0.5+cl, 1.0);
					glVertex3f(v[2].x, v[2].y, -v[2].z);
				glEnd();
			}
			else		/* intermediate bands */
			{
				v[0].x = ru * sind(b); v[0].y = upper_band; v[0].z = ru * cosd(b);
				n[0].x = v[0].x / r; n[0].y = v[0].y / r; n[0].z = v[0].z / r;
				v[1].x = ru * sind(b + step); v[1].y = upper_band; v[1].z = ru * cosd(b + step);
				n[1].x = v[1].x / r; n[1].y = v[1].y / r; n[1].z = v[1].z / r;
				v[2].x = rl * sind(b + step); v[2].y = lower_band; v[2].z = rl * cosd(b + step);
				n[2].x = v[2].x / r; n[2].y = v[2].y / r; n[2].z = v[2].z / r;
				v[3].x = rl * sind(b); v[3].y = lower_band; v[3].z = rl * cosd(b);
				n[3].x = v[3].x / r; n[3].y = v[3].y / r; n[3].z = v[3].z / r;
				glBegin(GL_QUADS);
					glColor3f(0.5+cu, 0.5+cu, 1.0);
					glVertex3f(v[0].x, v[0].y, -v[0].z);
					glColor3f(0.5+cl, 0.5+cl, 1.0);
					glVertex3f(v[3].x, v[3].y, -v[3].z);
					glColor3f(0.5+cl, 0.5+cl, 1.0);
					glVertex3f(v[2].x, v[2].y, -v[2].z);
					glColor3f(0.5+cu, 0.5+cu, 1.0);
					glVertex3f(v[1].x, v[1].y, -v[1].z);
				glEnd();
			}
		}
	}
	glEnable(GL_TEXTURE_2D);
	glPopMatrix();
}

/* Function to draw a terrain cell */

void draw_cell(terrain_cell *cell)
{
	terrain_cell *subcell;		/* sub-cell pointer */
	int sl, tb;					/* grid coords for bottom left corner of cell */
	int sc, tc;					/* grid coords for center of cell */
	int sr, tt;					/* grid coord for top right of cell */
	int subarrays, subarrayt;	/* sub-cell array coords */
	int fan_started = 0;		/* triangle fan started flag */

	/* set up grid coord data */

	sl = cell->s_datum;
	tb = cell->t_datum;
	sc = cell->s_datum + (cell->grid_step / 2);
	tc = cell->t_datum + (cell->grid_step / 2);
	sr = cell->s_datum + cell->grid_step;
	tt = cell->t_datum + cell->grid_step;
	subarrays = (cell->s_datum / cell->grid_step) * 2;
	subarrayt = (cell->t_datum / cell->grid_step) * 2;

	/* render bottom left quadrant */

	if (!(cell->flags & 0x0800))
	{
		fan_started = 1;
		glBegin(GL_TRIANGLE_FAN);
		glTexCoord2f((float)sc*tex_scale, (float)tc*tex_scale);
		glVertex3f((float)sc*GRID_FACTOR, (heightmap[sc][tc]*HEIGHT_FACTOR), -tc*GRID_FACTOR);
		if ((cell->flags & 0x0100) || (cell->flags & 0x0001))
		{
			glTexCoord2f((float)sl*tex_scale, (float)tc*tex_scale);
			glVertex3f((float)sl*GRID_FACTOR, (heightmap[sl][tc]*HEIGHT_FACTOR), -tc*GRID_FACTOR);
		}
		else
		{
			glTexCoord2f((float)sl*tex_scale, (float)tt*tex_scale);
			glVertex3f((float)sl*GRID_FACTOR, (heightmap[sl][tt]*HEIGHT_FACTOR), -tt*GRID_FACTOR);
		}
		glTexCoord2f((float)sl*tex_scale, (float)tb*tex_scale);
		glVertex3f((float)sl*GRID_FACTOR, (heightmap[sl][tb]*HEIGHT_FACTOR), -tb*GRID_FACTOR);
		if ((cell->flags & 0x0400) || (cell->flags & 0x0008))
		{
			glTexCoord2f((float)sc*tex_scale, (float)tb*tex_scale);
			glVertex3f((float)sc*GRID_FACTOR, (heightmap[sc][tb]*HEIGHT_FACTOR), -tb*GRID_FACTOR);
		}
	}
	else
	{
		switch (cell->lod)
		{
			case 5:
				subcell = &lod4[subarrays][subarrayt];
				break;
			case 4:
				subcell = &lod3[subarrays][subarrayt];
				break;
			case 3:
				subcell = &lod2[subarrays][subarrayt];
				break;
			case 2:
				subcell = &lod1[subarrays][subarrayt];
				break;
		}
		if (cell->lod > 1 && (subcell->flags & 0x8000))
		{
			draw_cell(subcell);
		}
	}

	/* render bottom right quadrant */

	if (!(cell->flags & 0x0400))
	{
		if (!fan_started)
		{
			fan_started = 1;
			glBegin(GL_TRIANGLE_FAN);
			glTexCoord2f((float)sc*tex_scale, (float)tc*tex_scale);
			glVertex3f((float)sc*GRID_FACTOR, (heightmap[sc][tc]*HEIGHT_FACTOR), -tc*GRID_FACTOR);
			if ((cell->flags & 0x0800) || (cell->flags & 0x0008))
			{
				glTexCoord2f((float)sc*tex_scale, (float)tb*tex_scale);
				glVertex3f((float)sc*GRID_FACTOR, (heightmap[sc][tb]*HEIGHT_FACTOR), -tb*GRID_FACTOR);
			}
		}
		glTexCoord2f((float)sr*tex_scale, (float)tb*tex_scale);
		glVertex3f((float)sr*GRID_FACTOR, (heightmap[sr][tb]*HEIGHT_FACTOR), -tb*GRID_FACTOR);
		if ((cell->flags & 0x0200) || (cell->flags & 0x0004))
		{
			glTexCoord2f((float)sr*tex_scale, (float)tc*tex_scale);
			glVertex3f((float)sr*GRID_FACTOR, (heightmap[sr][tc]*HEIGHT_FACTOR), -tc*GRID_FACTOR);
		}
	}
	else
	{
		glEnd();
		fan_started = 0;
		switch (cell->lod)
		{
			case 5:
				subcell = &lod4[subarrays+1][subarrayt];
				break;
			case 4:
				subcell = &lod3[subarrays+1][subarrayt];
				break;
			case 3:
				subcell = &lod2[subarrays+1][subarrayt];
				break;
			case 2:
				subcell = &lod1[subarrays+1][subarrayt];
				break;
		}
		if (cell->lod > 1 && (subcell->flags & 0x8000))
		{
			draw_cell(subcell);
		}
	}

	/* render top right quadrant */

	if (!(cell->flags & 0x0200))
	{
		if (!fan_started)
		{
			fan_started = 1;
			glBegin(GL_TRIANGLE_FAN);
			glTexCoord2f((float)sc*tex_scale, (float)tc*tex_scale);
			glVertex3f((float)sc*GRID_FACTOR, (heightmap[sc][tc]*HEIGHT_FACTOR), -tc*GRID_FACTOR);
			if ((cell->flags & 0x0400) || (cell->flags & 0x0004))
			{
				glTexCoord2f((float)sr*tex_scale, (float)tc*tex_scale);
				glVertex3f((float)sr*GRID_FACTOR, (heightmap[sr][tc]*HEIGHT_FACTOR), -tc*GRID_FACTOR);
			}
		}
		glTexCoord2f((float)sr*tex_scale, (float)tt*tex_scale);
		glVertex3f((float)sr*GRID_FACTOR, (heightmap[sr][tt]*HEIGHT_FACTOR), -tt*GRID_FACTOR);
		if ((cell->flags & 0x0100) || (cell->flags & 0x0002))
		{
			glTexCoord2f((float)sc*tex_scale, (float)tt*tex_scale);
			glVertex3f((float)sc*GRID_FACTOR, (heightmap[sc][tt]*HEIGHT_FACTOR), -tt*GRID_FACTOR);
		}
	}
	else
	{
		glEnd();
		fan_started = 0;
		switch (cell->lod)
		{
			case 5:
				subcell = &lod4[subarrays+1][subarrayt+1];
				break;
			case 4:
				subcell = &lod3[subarrays+1][subarrayt+1];
				break;
			case 3:
				subcell = &lod2[subarrays+1][subarrayt+1];
				break;
			case 2:
				subcell = &lod1[subarrays+1][subarrayt+1];
				break;
		}
		if (cell->lod > 1 && (subcell->flags & 0x8000))
		{
			draw_cell(subcell);
		}
	}

	/* render top left quadrant */

	if (!(cell->flags & 0x0100))
	{
		if (!fan_started)
		{
			fan_started = 1;
			glBegin(GL_TRIANGLE_FAN);
			glTexCoord2f((float)sc*tex_scale, (float)tc*tex_scale);
			glVertex3f((float)sc*GRID_FACTOR, (heightmap[sc][tc]*HEIGHT_FACTOR), -tc*GRID_FACTOR);
			if ((cell->flags & 0x0200) || (cell->flags & 0x0002))
			{
				glTexCoord2f((float)sc*tex_scale, (float)tt*tex_scale);
				glVertex3f((float)sc*GRID_FACTOR, (heightmap[sc][tt]*HEIGHT_FACTOR), -tt*GRID_FACTOR);
			}
		}
		glTexCoord2f((float)sl*tex_scale, (float)tt*tex_scale);
		glVertex3f((float)sl*GRID_FACTOR, (heightmap[sl][tt]*HEIGHT_FACTOR), -tt*GRID_FACTOR);
		if ((cell->flags & 0x0800) || (cell->flags & 0x0001))
		{
			glTexCoord2f((float)sl*tex_scale, (float)tc*tex_scale);
			glVertex3f((float)sl*GRID_FACTOR, (heightmap[sl][tc]*HEIGHT_FACTOR), -tc*GRID_FACTOR);
		}
	}
	else
	{
		glEnd();
		fan_started = 0;
		switch (cell->lod)
		{
			case 5:
				subcell = &lod4[subarrays][subarrayt+1];
				break;
			case 4:
				subcell = &lod3[subarrays][subarrayt+1];
				break;
			case 3:
				subcell = &lod2[subarrays][subarrayt+1];
				break;
			case 2:
				subcell = &lod1[subarrays][subarrayt+1];
				break;
		}
		if (cell->lod > 1 && (subcell->flags & 0x8000))
		{
			draw_cell(subcell);
		}
	}

	/* finish off fan if required */

	if (fan_started)
		glEnd();
}

