/* Landscape */
/* An OpenGL demo by A.P.Gardner */

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include <mmsystem.h>
#include "types.h"
#include "maths.h"
#include "terrain.h"

/* definitions */

#define MOVE_SPEED 2.0

/* variables */

GLuint texName[3];		/* store of texture references */
world_vector position;	/* camera position */
orientation angle;		/* camera orientation */
int left_mouse = 0;		/* indicates left mouse button is pressed */
int right_mouse = 0;	/* indicates right mouse button is pressed */
int mouse_x;			/* last position of mouse in x axis */
int mouse_y;			/* last position of mouse in y axis */
int fill = 1;			/* polygon fill mode flag */

long int start_time;
long int frame_count = 0;
long int last_update = 0;

/* Function to load a texture from a bmp file */

void make_texture_image(char filename[], int tex_ref)
{
	long data_start = 54;			/* start of data section in files wall.bmp & light.bmp */
									/* NB header size of bmp files is not fixed, refer to M$ documentation */
	unsigned char ch, cho[4];		/* temp file input buffers */
	FILE *fileptr;					/* pointer to bmp file */
	GLubyte image128[128][128][3];	/* storage for 128x128 texture */
	int n, m, i, j;					/* counters */

	/* open file */

	if ((fileptr = fopen(filename, "rb")) == NULL)
		return;

	/* step over the header */

	n = 0;
	while (!feof(fileptr) && n < data_start)
	{
		ch = fgetc(fileptr);
		n++;
	}

	/* load the texture data */

	for (i=0; i<128; i++)
	{
		for (j=0; j<128; j++)
		{
			m=0;
			while(!feof(fileptr) && m < 3)
			{
				cho[m] = fgetc(fileptr);
				m++;
			}
			image128[i][j][0] = cho[2];	/* bmp using BGR values, hence the swap */
			image128[i][j][1] = cho[1];
			image128[i][j][2] = cho[0];
		}
	}

	/* close the file */

	fclose(fileptr);

	/* create texture object */
	
	glBindTexture(GL_TEXTURE_2D, texName[tex_ref]);

	/* set up wrap & filter params */

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	/* load the image */

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 128, 128,
				 0, GL_RGB, GL_UNSIGNED_BYTE, image128);
}

/* Function to initialise params */

void init(void)
{
	GLfloat ambient_light[] = {0.5, 0.5, 0.5, 1.0};
	GLfloat position_one[] = {0.0, 0.9, 0.4, 0.0};
	GLfloat fogColor[] = {0.6, 0.6, 0.8, 1.0};

	/* initialise data */

	position.x = (GRID_FACTOR * 257.0) / 2.0; position.y = 20.0, position.z = 1.0;
	angle.pitch = 0.0; angle.yaw = 0.0; angle.roll = 0.0;

	/* generate texture names */

	texName[0] = 0;
	texName[1] = 0;
	texName[2] = 0;
	glGenTextures(3, texName);

	/* load textures */

	make_texture_image("ground.bmp", 0);
	make_texture_image("waves.bmp", 1);

	/* load terrain data and place viewer on surface */

	make_terrain("heightmap.bmp");
	position.y = find_height_above_terrain(position);

	/* set culling params */

	glEnable(GL_CULL_FACE);		
	glCullFace(GL_BACK);		/* cull back facing polygons */

	/* set clear colour */

	glClearColor(0.0, 0.0, 0.0, 1.0);

	/* enable depth buffering */

	glEnable(GL_DEPTH_TEST);

	/* set texturing function */

	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	/* set fog params */

	glFogfv(GL_FOG_COLOR, fogColor);
	glFogi(GL_FOG_MODE, GL_EXP2);
	glFogf(GL_FOG_DENSITY, 0.005);
}

/* Function to clean up */

void cleanup(void)
{
	/* delete texture objects */

	glDeleteTextures(3, texName);
}

/* Function to set camera position & direction */

void set_camera(GLfloat x, GLfloat y, GLfloat z,
			   GLfloat yaw, GLfloat pitch)
{
	/* rotate scene */

	glRotatef(-pitch, 1.0, 0.0, 0.0);
	glRotatef(yaw, 0.0, 1.0, 0.0);

	/* allow for translate scene */

	glTranslatef(-x, -y, z);
}

/* Function to draw display */

void display(void)
{
	/* clear all buffers */

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* clear the matrix */

	glLoadIdentity();

	/* set up the viewing transform */

	set_camera(position.x, position.y, position.z,
			   angle.yaw, angle.pitch);

	/* draw terrain */

	render_terrain();

	/* swap the front and back buffers */

	glutSwapBuffers();
}

/* Function to handle window size changes */

void reshape(int w, int h)
{
	glViewport(0, 0, (GLsizei)w, (GLsizei)h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0, (GLfloat)w/(GLfloat)h, 2.0, 500.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

/* Function to handle key presses */

void keyboard(unsigned char key, int x, int y)
{
	world_vector dir;

	if (key == 's' || key == 'S')
	{
		dir = orientation_to_unit_vector(angle);
		dir = world_vector_scale(dir, -MOVE_SPEED);
		position = world_vector_add(position, dir);
		position.y = find_height_above_terrain(position);
	}
	
	if (key == 'f' || key == 'F')
	{
		if (fill)
		{
			fill = 0;
			glPolygonMode(GL_FRONT, GL_LINE);
		}
		else
		{
			fill = 1;
			glPolygonMode(GL_FRONT, GL_FILL);
		}
	}		

}

/* Function to handle mouse inputs */

void mouse(int button, int state, int x, int y)
{
	/* check for mouse button state changes */

	if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
		right_mouse = 1;

	if (button == GLUT_RIGHT_BUTTON && state == GLUT_UP)
		right_mouse = 0;

	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
		left_mouse = 1;

	if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
		left_mouse = 0;

	/* store mouse position */

	mouse_x = x;
	mouse_y = y;
}

/* Function to handle mouse movements */

void motion(int x, int y)
{
	float mouse_move_x = 0.0, mouse_move_y = 0.0;

	/* check for movements */

	if (x != mouse_x)
	{
		mouse_move_x = x - mouse_x;
		mouse_x = x;
	}

	if (y != mouse_y)
	{
		mouse_move_y = y - mouse_y;
		mouse_y = y;
	}

	/* change orientation of camera */

	if (mouse_move_x != 0.0)
	{
		angle.yaw = angle.yaw + mouse_move_x;
		if (angle.yaw > 180.0)
		{
			angle.yaw = angle.yaw - 360.0;
		}
		if (angle.yaw < -180.0)
		{
			angle.yaw = angle.yaw + 360.0;
		}
	}

	if (mouse_move_y != 0.0)
	{
		angle.pitch = angle.pitch + mouse_move_y;
		if (angle.pitch > 90.0)
		{
			angle.pitch = 90.0;
		}
		if (angle.pitch < -90.0)
		{
			angle.pitch = -90.0;
		}
	}
}

/* Idle function */

void idle(void)
{
	world_vector dir;

	/* move camera if required */

	if (right_mouse)
	{
		dir = orientation_to_unit_vector(angle);
		dir = world_vector_scale(dir, MOVE_SPEED);
		position = world_vector_add(position, dir);
		position.y = find_height_above_terrain(position);
	}		

	/* update display */

	glutPostRedisplay();
}

/* Main function */

int main(int argc, char** argv)
{
	/* print instructions */

	printf("LANDSCAPE DEMO\n\n");
	printf("Controls:\n\n");
	printf("Left mouse button - mouse movements change view\n");
	printf("Right mouse button - move forward\n");
	printf("S - move backward\n");
	printf("F - toggle fill mode\n");

	/* run the demo */

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGB);
	glutInitWindowSize(640, 480);
	glutInitWindowPosition(100, 100);
	glutCreateWindow("OpenGL Demo - Landscape");
	init();
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutIdleFunc(idle);
	glutMainLoop();
	cleanup();

	return 0;
}