/* File:		make_terrain.c */
/* Description:	This file module contains the terrain pre-processing functions */

#include <stdio.h>
#include <math.h>
#include <GL/glut.h>
#include "types.h"
#include "maths.h"
#include "terrain.h"

/* variables */

extern GLuint texName[2];

/* Function to pre-process heightmap to provide better water areas */

void adjust_water_vertices()
{
	int i, j;
	int water, land;

	for (i=0; i<255; i++)
	{
		for (j=0; j<255; j++)
		{
			/* determine type of tile */

			water = 0;
			land = 0;
			if (heightmap[i][j] > WATER_HEIGHT)
			{
				land++;
			}
			else
			{
				water++;
			}
			if (heightmap[i+1][j] > WATER_HEIGHT)
			{
				land++;
			}
			else
			{
				water++;
			}
			if (heightmap[i][j+1] > WATER_HEIGHT)
			{
				land++;
			}
			else
			{
				water++;
			}
			if (heightmap[i+1][j+1] > WATER_HEIGHT)
			{
				land++;
			}
			else
			{
				water++;
			}

			/* process tile */

			if (water == 4)
			{
				heightmap[i][j] = heightmap[i][j] / 4;
				heightmap[i+1][j] = heightmap[i][j] / 4;
				heightmap[i][j+1] = heightmap[i][j] / 4;
				heightmap[i+1][j+1] = heightmap[i][j] / 4;
			}
			else if (water > 0)
			{
				if (heightmap[i][j] < WATER_HEIGHT)
				{
					heightmap[i][j] = WATER_HEIGHT;
				}
				if (heightmap[i+1][j] < WATER_HEIGHT)
				{
					heightmap[i+1][j] = WATER_HEIGHT;
				}
				if (heightmap[i][j+1] < WATER_HEIGHT)
				{
					heightmap[i][j+1] = WATER_HEIGHT;
				}
				if (heightmap[i+1][j+1] < WATER_HEIGHT)
				{
					heightmap[i+1][j+1] = WATER_HEIGHT;
				}
			}
		}
	}
}

/* Function to calculate a colour value for a vertex */

colour_rgb calculate_vertex_colour(unsigned char height)
{
	colour_rgb c;
	float alpha;

	if (height <= WATER_HEIGHT)
	{
		c.r = WATER_COLOUR_R;
		c.g = WATER_COLOUR_G;
		c.b = WATER_COLOUR_B;
	} else if (height < SHORE_HEIGHT)
	{
		alpha = (float)(height - WATER_HEIGHT) / (float)(SHORE_HEIGHT - WATER_HEIGHT);
		c.r = ((1.0 - alpha) * WATER_COLOUR_R) + (alpha * GRASS_COLOUR_R);
		c.g = ((1.0 - alpha) * WATER_COLOUR_G) + (alpha * GRASS_COLOUR_G);
		c.b = ((1.0 - alpha) * WATER_COLOUR_B) + (alpha * GRASS_COLOUR_B);
	} else if (height < GRASS_HEIGHT)
	{
		c.r = GRASS_COLOUR_R;
		c.g = GRASS_COLOUR_G;
		c.b = GRASS_COLOUR_B;
	} else if (height < ROCK_HEIGHT)
	{
		alpha = (float)(height - GRASS_HEIGHT) / (float)(ROCK_HEIGHT - GRASS_HEIGHT);
		c.r = ((1.0 - alpha) * GRASS_COLOUR_R) + (alpha * SNOW_COLOUR_R);
		c.g = ((1.0 - alpha) * GRASS_COLOUR_G) + (alpha * SNOW_COLOUR_G);
		c.b = ((1.0 - alpha) * GRASS_COLOUR_B) + (alpha * SNOW_COLOUR_B);
	} else
	{
		c.r = SNOW_COLOUR_R;
		c.g = SNOW_COLOUR_G;
		c.b = SNOW_COLOUR_B;
	}

	return c;
}

/* Function to calculate the lighting for the vertices */

void calculate_vertex_lighting()
{
	world_vector v1, v2, v3;			/* temporary vertex stores */
	world_vector n;						/* current vertex normal */
	int num_normals;					/* number of quad normals that apply to a vertex normal */
	world_vector lv;					/* direction of light source */
	float intensity;					/* calculated light intensity for current vertex */
	int x, z;							/* counters */

	/* calculate the normal for each quad panel in the scene */

	for (x=0; x<256; x++)
	{
		for (z=0; z<256; z++)
		{
			v1.x = x*GRID_FACTOR; v1.y = heightmap[x][z]*HEIGHT_FACTOR; v1.z = z*GRID_FACTOR;
			v2.x = x*GRID_FACTOR; v2.y = heightmap[x][z+1]*HEIGHT_FACTOR; v2.z = (z+1)*GRID_FACTOR;
			v3.x = (x+1)*GRID_FACTOR; v3.y = heightmap[x+1][z+1]*HEIGHT_FACTOR; v3.z = (z+1)*GRID_FACTOR;
			normalmap[x][z] = calculate_panel_normal(v1, v2, v3);
		}
	}

	/* average the panel normals at each vertex and calculate the lighting */

	for (x=0; x<257; x++)
	{
		for (z=0; z<257; z++)
		{
			num_normals = 0;
			n.x = 0.0; n.y = 0.0; n.z = 0.0;
			if (x>0 && z<256)
			{
				n.x = n.x + normalmap[x-1][z].x;
				n.y = n.y + normalmap[x-1][z].y;
				n.z = n.z + normalmap[x-1][z].z;
				num_normals++;
			}
			if (x<256 && z<256)
			{
				n.x = n.x + normalmap[x][z].x;
				n.y = n.y + normalmap[x][z].y;
				n.z = n.z + normalmap[x][z].z;
				num_normals++;
			}
			if (x<256 && z>0)
			{
				n.x = n.x + normalmap[x][z-1].x;
				n.y = n.y + normalmap[x][z-1].y;
				n.z = n.z + normalmap[x][z-1].z;
				num_normals++;
			}
			if (x>0 && z>0)
			{
				n.x = n.x + normalmap[x-1][z-1].x;
				n.y = n.y + normalmap[x-1][z-1].y;
				n.z = n.z + normalmap[x-1][z-1].z;
				num_normals++;
			}

			/* calculate vertex normal */

			n = world_vector_scale(n, (1.0/(float)num_normals));

			/* apply lighting */

			lv.x = LIGHT_ANGLE_X; lv.y = LIGHT_ANGLE_Y; lv.z = LIGHT_ANGLE_Z;
			intensity = -world_vector_dot(n, lv);
			if (intensity < 0.0)
			{
				intensity = 0.0;
			}
			colourmap[x][z].r = colourmap[x][z].r * (AMBIENT_LIGHT_R + (intensity * LIGHT_COLOUR_R));
			colourmap[x][z].g = colourmap[x][z].g * (AMBIENT_LIGHT_G + (intensity * LIGHT_COLOUR_G));
			colourmap[x][z].b = colourmap[x][z].b * (AMBIENT_LIGHT_B + (intensity * LIGHT_COLOUR_B));
		}
	}

}

/* Function to make the terrain texture from the colour map */

void make_terrain_texture()
{
	GLubyte image[256][256][3];		/* storage for 256x256 texture */
	int i, j;						/* counters */

	/* load the texture data */

	for (i=0; i<256; i++)
	{
		for (j=0; j<256; j++)
		{
			image[i][j][0] = (GLubyte)(colourmap[j][i].r * 256);
			image[i][j][1] = (GLubyte)(colourmap[j][i].g * 256);
			image[i][j][2] = (GLubyte)(colourmap[j][i].b * 256);
		}
	}

	/* create texture object */
	
	glBindTexture(GL_TEXTURE_2D, texName[2]);

	/* set up wrap & filter params */

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	/* load the image */

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, 256, 256,
				 0, GL_RGB, GL_UNSIGNED_BYTE, image);
}

/* Function to initialise LOD data arrays */

void initialise_lod_data()
{
	int n, m;
	unsigned short s0, t0, step, mid;

	step = pow(2, 5);
	mid = step / 2;
	for (n=0; n<8; n++)
	{
		for (m=0; m<8; m++)
		{
			s0 = n * step;
			t0 = m * step;
			lod5[n][m].lod = 5;
			lod5[n][m].grid_step = step;
			lod5[n][m].s_datum = s0;
			lod5[n][m].t_datum = t0;
			lod5[n][m].flags = 0;
			lod5[n][m].variance[0] = abs(heightmap[s0+mid][t0]-
									((heightmap[s0][t0]+heightmap[s0+step][t0])/2));
			lod5[n][m].variance[1] = abs(heightmap[s0+step][t0+mid]-
									((heightmap[s0+step][t0]+heightmap[s0+step][t0+step])/2));
			lod5[n][m].variance[2] = abs(heightmap[s0+mid][t0+step]-
									((heightmap[s0+step][t0+step]+heightmap[s0][t0+step])/2));
			lod5[n][m].variance[3] = abs(heightmap[s0][t0+mid]-
									((heightmap[s0][t0+step]+heightmap[s0][t0])/2));
		}
	}
	step = pow(2, 4);
	mid = step / 2;
	for (n=0; n<16; n++)
	{
		for (m=0; m<16; m++)
		{
			s0 = n * step;
			t0 = m * step;
			lod4[n][m].lod = 4;
			lod4[n][m].grid_step = step;
			lod4[n][m].s_datum = s0;
			lod4[n][m].t_datum = t0;
			lod4[n][m].flags = 0;
			lod4[n][m].variance[0] = abs(heightmap[s0+mid][t0]-
									((heightmap[s0][t0]+heightmap[s0+step][t0])/2));
			lod4[n][m].variance[1] = abs(heightmap[s0+step][t0+mid]-
									((heightmap[s0+step][t0]+heightmap[s0+step][t0+step])/2));
			lod4[n][m].variance[2] = abs(heightmap[s0+mid][t0+step]-
									((heightmap[s0+step][t0+step]+heightmap[s0][t0+step])/2));
			lod4[n][m].variance[3] = abs(heightmap[s0][t0+mid]-
									((heightmap[s0][t0+step]+heightmap[s0][t0])/2));
		}
	}
	step = pow(2, 3);
	mid = step / 2;
	for (n=0; n<32; n++)
	{
		for (m=0; m<32; m++)
		{
			s0 = n * step;
			t0 = m * step;
			lod3[n][m].lod = 3;
			lod3[n][m].grid_step = step;
			lod3[n][m].s_datum = s0;
			lod3[n][m].t_datum = t0;
			lod3[n][m].flags = 0;
			lod3[n][m].variance[0] = abs(heightmap[s0+mid][t0]-
									((heightmap[s0][t0]+heightmap[s0+step][t0])/2));
			lod3[n][m].variance[1] = abs(heightmap[s0+step][t0+mid]-
									((heightmap[s0+step][t0]+heightmap[s0+step][t0+step])/2));
			lod3[n][m].variance[2] = abs(heightmap[s0+mid][t0+step]-
									((heightmap[s0+step][t0+step]+heightmap[s0][t0+step])/2));
			lod3[n][m].variance[3] = abs(heightmap[s0][t0+mid]-
									((heightmap[s0][t0+step]+heightmap[s0][t0])/2));
		}
	}
	step = pow(2, 2);
	mid = step / 2;
	for (n=0; n<64; n++)
	{
		for (m=0; m<64; m++)
		{
			s0 = n * step;
			t0 = m * step;
			lod2[n][m].lod = 2;
			lod2[n][m].grid_step = step;
			lod2[n][m].s_datum = s0;
			lod2[n][m].t_datum = t0;
			lod2[n][m].flags = 0;
			lod2[n][m].variance[0] = abs(heightmap[s0+mid][t0]-
									((heightmap[s0][t0]+heightmap[s0+step][t0])/2));
			lod2[n][m].variance[1] = abs(heightmap[s0+step][t0+mid]-
									((heightmap[s0+step][t0]+heightmap[s0+step][t0+step])/2));
			lod2[n][m].variance[2] = abs(heightmap[s0+mid][t0+step]-
									((heightmap[s0+step][t0+step]+heightmap[s0][t0+step])/2));
			lod2[n][m].variance[3] = abs(heightmap[s0][t0+mid]-
									((heightmap[s0][t0+step]+heightmap[s0][t0])/2));
		}
	}
	step = pow(2, 1);
	mid = step / 2;
	for (n=0; n<128; n++)
	{
		for (m=0; m<128; m++)
		{
			s0 = n * step;
			t0 = m * step;
			lod1[n][m].lod = 1;
			lod1[n][m].grid_step = step;
			lod1[n][m].s_datum = s0;
			lod1[n][m].t_datum = t0;
			lod1[n][m].flags = 0;
			lod1[n][m].variance[0] = abs(heightmap[s0+mid][t0]-
									((heightmap[s0][t0]+heightmap[s0+step][t0])/2));
			lod1[n][m].variance[1] = abs(heightmap[s0+step][t0+mid]-
									((heightmap[s0+step][t0]+heightmap[s0+step][t0+step])/2));
			lod1[n][m].variance[2] = abs(heightmap[s0+mid][t0+step]-
									((heightmap[s0+step][t0+step]+heightmap[s0][t0+step])/2));
			lod1[n][m].variance[3] = abs(heightmap[s0][t0+mid]-
									((heightmap[s0][t0+step]+heightmap[s0][t0])/2));
		}
	}
}

