/* File:		terrain.c */
/* Description:	This file module contains the terrain loading and rendering functions */

#include <stdio.h>
#include <math.h>
#include <GL/glut.h>
#include "types.h"
#include "maths.h"
#include "terrain.h"
#include "make_terrain.h"
#include "draw_terrain.h"

#define FAR_CLIP 500.0
#define PLAYER_HEIGHT 6.0

/* variables */

extern GLuint texName[2];
extern world_vector position;
extern orientation angle;
unsigned char heightmap[257][257];
world_vector normalmap[256][256];	/* array of normals for each quad in grid */
colour_rgb colourmap[257][257];
grid_vector distancemap[257][257];
terrain_cell lod5[8][8];
terrain_cell lod4[16][16];
terrain_cell lod3[32][32];
terrain_cell lod2[64][64];
terrain_cell lod1[128][128];
int detail_level = 1;
float lod_factor;
GLint gl_max_texture_size;
float tex_scale;

/* Function to give bottom-left grid coordinate for a given position */

void position_to_grid_coord(world_vector pos, int* s, int* t)
{
	*s = (int)(pos.x / GRID_FACTOR);
	*t = (int)(pos.z / GRID_FACTOR);
}

/* Function to find center of a grid tile given its bottom-left grid coordinate */

void grid_coord_to_center_position(int s, int t, world_vector* pos)
{
	pos->x = ((float)s * GRID_FACTOR) + 0.5;
	pos->y = 0.0;
	pos->z = ((float)t * GRID_FACTOR) + 0.5;
}

/* Function to determine vertical distance and intercept for a grid square */

void calculate_grid_intercept(world_vector point, int s, int t,
							  world_vector *intercept, float *range)
{
	int n;
	int panel_hit = 1;
	world_vector v1, va, vb;
	float ndotp, plane_constant, ndotv;
	world_vector acrossb;
	world_vector vertex;
	world_vector v = {0.0, -1.0, 0.0};

	/* initialise return values */

	*range = FAR_CLIP * 10.0;
	intercept->x = 0.0; intercept->y = 0.0; intercept->z = 0.0;

	/* find ndotv and check it */

	ndotv = world_vector_dot(normalmap[s][t], v);

	/* calculate range to panel's plane */

	vertex.x = (GRID_FACTOR*(float)s);
	vertex.y = heightmap[s][t]*HEIGHT_FACTOR;
	vertex.z = (GRID_FACTOR*(float)t);
	ndotp = world_vector_dot(normalmap[s][t], point);
	plane_constant = world_vector_dot(normalmap[s][t], vertex);
	*range = (plane_constant - ndotp) / ndotv;

	/* calculate intercept point with panel's plane */

	v1.x = *range * v.x; v1.y = *range * v.y; v1.z = *range * v.z;
	*intercept = world_vector_add(point, v1);
}

/* Function to give new height above scene */

float find_height_above_terrain(world_vector p)
{
	int s, t;
	world_vector intercept;
	float range;

	position_to_grid_coord(p, &s, &t);
	calculate_grid_intercept(p, s, t, &intercept, &range);
	return intercept.y + PLAYER_HEIGHT - 0.5;
}

/************************ make terrain ***********************************/

/* Function to load and create a terrain database from a bmp file */

void make_terrain(char filename[])
{
	long data_start = 54;		/* start of data section in terrain .bmp files */
								/* NB header size of bmp files is not fixed, refer to M$ documentation */
	unsigned char ch, cho[3];	/* temp file input buffers */
	FILE *fileptr, *fileptr2;				/* pointer to bmp file */
	int n, m, i, j;				/* counters */

	/* determine system params */

	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &gl_max_texture_size);

	/* open file */

	if ((fileptr = fopen(filename, "rb")) == NULL)
		return;

	/* step over the header */

	n = 0;
	while (!feof(fileptr) && n < data_start)
	{
		ch = fgetc(fileptr);
		n++;
	}

	/* load the height map data */

	for (i=0; i<257; i++)
	{
		for (j=0; j<257; j++)
		{
			cho[0] = fgetc(fileptr);
			cho[1] = fgetc(fileptr);
			cho[2] = fgetc(fileptr);
			heightmap[j][i] = cho[0];
		}
		cho[0] = fgetc(fileptr);	/* remove padding byte */
	}

	/* close the file */

	fclose(fileptr);

	/* pre-process the height map */

	adjust_water_vertices();

	/* generate the colour map data */

	for (i=0; i<257; i++)
	{
		for (j=0; j<257; j++)
		{
			colourmap[i][j] = calculate_vertex_colour(heightmap[i][j]);
		}
	}

	/* light the scene */

	calculate_vertex_lighting();

	/* make terrain texture */

	make_terrain_texture();
	tex_scale = 1.0 / 256.0;

	/* generate the LOD data */

	lod_factor = LOD_FACTOR * (float)detail_level;
	initialise_lod_data();
}

/*************************** render terrain ********************************/

/* Function to render terrain */

void render_terrain()
{
	int s, t;
	terrain_cell *cell;

	/* calculate grid distances relative to viewer for all grid squares */

	calculate_distance_map();

	/* calculate lod for current view */

	cull_cells();
	calculate_lod();

	/* render sky */

	draw_sky();

	/* render terrain */

	glEnable(GL_FOG);
	glColor3f(1.0, 1.0, 1.0);
	glBindTexture(GL_TEXTURE_2D, texName[2]);
	for (t=0; t<8; t++)
	{
		for (s=0; s<8; s++)
		{
			cell = &lod5[s][t];
			if (cell->flags & 0x8000)
			{
				draw_cell(cell);
			}
		}
	}
	glDisable(GL_FOG);
	
	/* render water */

	glBindTexture(GL_TEXTURE_2D, texName[1]);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
	glColor3f(1.0, 1.0, 1.0);
	glBegin(GL_QUADS);
		glTexCoord2f(0.0, 256*WATER_TEX_SCALE);
		glVertex3f(0.0, ((WATER_HEIGHT-0.1)*HEIGHT_FACTOR), -(256*GRID_FACTOR));
		glTexCoord2f(0.0, 0.0);
		glVertex3f(0.0, ((WATER_HEIGHT-0.1)*HEIGHT_FACTOR), 0.0);
		glTexCoord2f(256*WATER_TEX_SCALE, 0.0);
		glVertex3f((256*GRID_FACTOR), ((WATER_HEIGHT-0.1)*HEIGHT_FACTOR), 0.0);
		glTexCoord2f(256*WATER_TEX_SCALE, 256*WATER_TEX_SCALE);
		glVertex3f((256*GRID_FACTOR), ((WATER_HEIGHT-0.1)*HEIGHT_FACTOR), -(256*GRID_FACTOR));
	glEnd();
	glDisable(GL_BLEND);

}
